# -*- coding: utf-8 -*-

import asyncio
import json

from asgiref.sync import sync_to_async

import django
from django.conf import settings
from django.core.handlers.asgi import ASGIHandler, ASGIRequest
from django.core import signals
from django.http import QueryDict
from django.test.utils import override_settings
from django.urls import get_resolver, set_script_prefix, set_urlconf

from pistoke.ohjain import WEBSOCKET_MIDDLEWARE


class WebsocketPyynto(ASGIRequest):
  '''
  Yksittäisen Websocket-pyynnön (istunnon) tiedot.

  Huomaa, että __init__ ei kutsu super-toteutusta.
  '''
  # pylint: disable=too-many-instance-attributes
  # pylint: disable=method-hidden
  # pylint: disable=invalid-name

  POST = QueryDict()
  FILES = {}

  def __init__(self, scope):
    # pylint: disable=super-init-not-called
    self.scope = scope
    self._post_parse_error = False
    self._read_started = False
    self.resolver_match = None
    self.script_name = self.scope.get('root_path', '')
    if self.script_name and scope['path'].startswith(self.script_name):
      self.path_info = scope['path'][len(self.script_name):]
    else:
      self.path_info = scope['path']
    if self.script_name:
      self.path = '%s/%s' % (
        self.script_name.rstrip('/'),
        self.path_info.replace('/', '', 1),
      )
    else:
      self.path = scope['path']

    self.method = 'Websocket'

    query_string = self.scope.get('query_string', '')
    if isinstance(query_string, bytes):
      query_string = query_string.decode()
    self.META = {
      'REQUEST_METHOD': self.method,
      'QUERY_STRING': query_string,
      'SCRIPT_NAME': self.script_name,
      'PATH_INFO': self.path_info,
      'wsgi.multithread': True,
      'wsgi.multiprocess': True,
    }
    if self.scope.get('client'):
      self.META['REMOTE_ADDR'] = self.scope['client'][0]
      self.META['REMOTE_HOST'] = self.META['REMOTE_ADDR']
      self.META['REMOTE_PORT'] = self.scope['client'][1]
    if self.scope.get('server'):
      self.META['SERVER_NAME'] = self.scope['server'][0]
      self.META['SERVER_PORT'] = str(self.scope['server'][1])
    else:
      self.META['SERVER_NAME'] = 'unknown'
      self.META['SERVER_PORT'] = '0'
    for name, value in self.scope.get('headers', []):
      name = name.decode('latin1')
      corrected_name = 'HTTP_%s' % name.upper().replace('-', '_')
      value = value.decode('latin1')
      if corrected_name in self.META:
        value = self.META[corrected_name] + ',' + value
      self.META[corrected_name] = value

    self.resolver_match = None
    # def __init__

  # class WebsocketPyynto


class WebsocketKasittelija(ASGIHandler):
  '''
  Saapuvien Websocket-pyyntöjen (istuntojen) käsittelyrutiini.
  '''
  def __new__(cls, *args, **kwargs):
    '''
    Alusta Django ennen käsittelyrutiinin luontia.

    Vrt. get_asgi_application().
    '''
    django.setup(set_prefix=False)
    return super().__new__(cls, *args, **kwargs)
    # def __new__

  async def __call__(self, scope, receive, send):
    '''
    Asynkroninen, pyyntökohtainen kutsu.

    Vrt. django.core.handlers.asgi:ASGIHandler.__call__
    '''
    assert scope['type'] == 'websocket'

    set_script_prefix(self.get_script_prefix(scope))
    signals.request_started.send(sender=self.__class__, scope=scope)

    # Suorita Websocket-kättely.
    while (await receive())['type'] != 'websocket.connect':
      pass
    await send({'type': 'websocket.accept'})

    # Muodosta pyyntö.
    request = WebsocketPyynto(scope)

    # Hae käsittelevä näkymärutiini tai mahdollinen virheviesti.
    if asyncio.iscoroutinefunction(self.get_response):
      nakyma = await self.get_response(request)
    else:
      nakyma = await sync_to_async(self.get_response)(request)

    # Palauta virhesanoma, mikäli `dispatch` tuotti sellaisen.
    if not asyncio.iscoroutine(nakyma):
      return await self.send_response(nakyma, send)

    # Luodaan asynkroninen tehtävä näkymän suorituksesta.
    nakyma_tehtava = asyncio.ensure_future(nakyma)

    # Luodaan jono saapuville syötteille.
    _syote = asyncio.Queue()
    async def syote():
      '''
      Erillinen metodi saapuvan syötteen käsittelyyn.

      Poimitaan ja toteutetaan mahdollinen katkaisupyyntö
      riippumatta siitä, lukeeko näkymärutiini syötettä vai ei.
      '''
      while True:
        sanoma = await receive()
        if sanoma['type'] == 'websocket.receive':
          await _syote.put(sanoma.get('text', sanoma.get('bytes', None)))
        elif sanoma['type'] == 'websocket.disconnect':
          break
        else:
          raise ValueError
        # while True
      # async def syote
    syote_tehtava = asyncio.ensure_future(syote())

    async def _send(data):
      '''
      Lähetetään annettu data joko tekstinä tai tavujonona.
      '''
      if isinstance(data, str):
        return await send({'type': 'websocket.send', 'text': data})
      elif isinstance(data, bytearray):
        return await send({'type': 'websocket.send', 'bytes': bytes(data)})
      elif isinstance(data, bytes):
        return await send({'type': 'websocket.send', 'bytes': data})
      # async def _send

    # pylint: disable=attribute-defined-outside-init
    # Kääri ASGI-protokollan mukaiset `receive`- ja `send`-metodit
    # Websocket-metodeiksi.
    request.receive = _syote.get
    request.send = _send
    # pylint: enable=attribute-defined-outside-init

    # Alusta tehtävät.
    tehtavat = {syote_tehtava, nakyma_tehtava}

    # Odota siksi kunnes joko syöte katkaistaan
    # tai näkymärutiini on valmis.
    try:
      _, tehtavat = await asyncio.wait(
        tehtavat, return_when=asyncio.FIRST_COMPLETED
      )

    # Peruuta kesken jääneet tehtävät ja odota ne loppuun.
    finally:
      for kesken in tehtavat:
        kesken.cancel()
      await asyncio.gather(*tehtavat, return_exceptions=True)
    # async def __call__

  def load_middleware(self):
    '''
    Ajetaan vain muunnostaulun mukaan sallitut ohjaimet.
    '''
    with override_settings(MIDDLEWARE=list(filter(None, (
      ws_ohjain if isinstance(ws_ohjain, str)
      else ohjain if ws_ohjain else None
      for ohjain, ws_ohjain in (
        (ohjain, WEBSOCKET_MIDDLEWARE.get(ohjain, False))
        for ohjain in settings.MIDDLEWARE
      )
    )))):
      super().load_middleware()
    # def load_middleware

  def get_response(self, request):
    '''
    Ohitetaan paluusanoman käsittelyyn liittyvät funktiokutsut.
    '''
    set_urlconf(settings.ROOT_URLCONF)
    return self._middleware_chain(request)
    # def get_response

  def _get_response(self, request):
    '''
    Ohitetaan paluusanoman käsittelyyn liittyvät funktiokutsut.
    '''
    if hasattr(request, 'urlconf'):
      urlconf = request.urlconf
      set_urlconf(urlconf)
      resolver = get_resolver(urlconf)
    else:
      resolver = get_resolver()

    resolver_match = resolver.resolve(request.path_info)
    callback, callback_args, callback_kwargs = resolver_match
    request.resolver_match = resolver_match

    for middleware_method in self._view_middleware:
      middleware_method(request, callback, callback_args, callback_kwargs)

    # pylint: disable=not-callable
    return callback(request, *callback_args, **callback_kwargs)
    # def _get_response

  async def send_response(self, response, send):
    '''
    Lähetä HTTP-virhesanoma soveltuvin osin
    Websocket-prokollan mukaisesti.

    Vrt. super-toteutus.
    '''
    # pylint: disable=invalid-name
    # Collect cookies into headers. Have to preserve header case as there
    # are some non-RFC compliant clients that require e.g. Content-Type.
    response_headers = []
    for header, value in response.items():
      response_headers.append((header, value))
    for c in response.cookies.values():
      response_headers.append(
        ('Set-Cookie', c.output(header='').strip())
      )
    # Initial response message.
    await send({
      'type': 'websocket.send',
      'text': json.dumps({
        'status': response.status_code,
        'headers': response_headers,
      }),
    })
    # Streaming responses need to be pinned to their iterator.
    if response.streaming:
      # Access `__iter__` and not `streaming_content` directly in case
      # it has been overridden in a subclass.
      for part in response:
        for chunk, _ in self.chunk_bytes(part):
          await send({
            'type': 'websocket.send',
            'bytes': chunk,
          })
    # Other responses just need chunking.
    else:
      # Yield chunks of response.
      for chunk, _ in self.chunk_bytes(response.content):
        await send({
          'type': 'websocket.send',
          'bytes': chunk,
        })
    response.close()
    # async def send_response

  # class WebsocketKasittelija



# -*- coding: utf-8 -*-

from django.contrib.sessions.middleware import SessionMiddleware
from django.core.handlers.asgi import ASGIRequest
from django.middleware.csrf import CsrfViewMiddleware


class WebsocketOhjain:
  '''
  Ohjain, joka asettaa pyynnölle URI-määritteen `websocket` siten,
  että siihen tehdyt pyynnöt ohjataan ajossa olevaan ASGI-käsittelijään.

  Mikäli HTTP-yhteys on salattu (HTTPS),
  käytetään salattua Websocket-yhteyttä (WSS).

  Mikäli pyyntö ei ole ASGI-pohjainen, ei `websocket`-yhteys ole
  käytettävissä eikä em. määritettä aseteta.
  '''
  def __init__(self, get_response):
    self.get_response = get_response

  def __call__(self, request):
    if isinstance(request, ASGIRequest):
      request.websocket = (
        f'{"wss" if request.is_secure() else "ws"}://{request.get_host()}'
      )
    return self.get_response(request)
    # def __call__

  # class WebsocketOhjain


class OhitaPaluusanoma:
  ''' Saateluokka, joka ohittaa paluusanoman käsittelyn. '''
  def process_response(self, request, response):
    return response


class CsrfOhjain(OhitaPaluusanoma, CsrfViewMiddleware):
  def process_view(self, request, callback, callback_args, callback_kwargs):
    '''
    Lisätään Websocket-pyyntöön metodi `_tarkista_csrf` pyyntödatan
    mukana luettavan CSRF-tunnisteen tarkistamiseen.

    Ohitetaan tavanomainen CSRF-tarkistus POST-datasta (super).
    '''
    # pylint: disable=protected-access
    from django.middleware.csrf import (
      _compare_salted_tokens,
      _sanitize_token,
    )
    def _tarkista_csrf(csrf_token):
      return _compare_salted_tokens(
        _sanitize_token(csrf_token), request.META.get('CSRF_COOKIE')
      )
    request._tarkista_csrf = _tarkista_csrf
    # def process_view
  # class CsrfOhjain


class IstuntoOhjain(OhitaPaluusanoma, SessionMiddleware):
  pass


# Muunnostaulu Websocket-pyyntöihin sovellettavista
# Middleware-ohjaimista.
# Muiden kuin tässä mainittujen ohjaimien lataus ohitetaan.
WEBSOCKET_MIDDLEWARE = {
  # Ohitetaan.
  'corsheaders.middleware.CorsMiddleware': False,
  'debug_toolbar.middleware.DebugToolbarMiddleware': False,
  'django.middleware.gzip.GZipMiddleware': False,
  'django.middleware.security.SecurityMiddleware' : False,
  'django.middleware.common.CommonMiddleware': False,
  'django.contrib.messages.middleware.MessageMiddleware': False,
  'django.middleware.clickjacking.XFrameOptionsMiddleware': False,
  'django_hosts.middleware.HostsResponseMiddleware': False,

  # Suoritetaan.
  'django_hosts.middleware.HostsRequestMiddleware': True,
  'django.middleware.csrf.CsrfViewMiddleware': 'pistoke.ohjain.CsrfOhjain',
  'django.contrib.sessions.middleware.SessionMiddleware': \
    'pistoke.ohjain.IstuntoOhjain',
  'django.contrib.auth.middleware.AuthenticationMiddleware': True,
  'impersonate.middleware.ImpersonateMiddleware': True,
  'silk.middleware.SilkyMiddleware': True,
  'pistoke.ohjain.WebsocketOhjain': True,
}



# -*- coding: utf-8 -*-

import setuptools

setuptools._install_setup_requires({'setup_requires': ['git-versiointi']})
from versiointi import asennustiedot

setuptools.setup(
  name='django-pistoke',
  description='Django-Websocket-laajennos',
  url='https://github.com/an7oine/django-pistoke.git',
  author='Antti Hautaniemi',
  author_email='antti.hautaniemi@pispalanit.fi',
  packages=setuptools.find_packages(),
  include_package_data=True,
  zip_safe=False,
  extras_require={
    'runserver': ['uvicorn'],
  },
  **asennustiedot(__file__),
)



# -*- coding: utf-8 -*-

# pylint: disable=invalid-name
# pylint: disable=ungrouped-imports

from urllib.parse import urlparse

from django.conf import settings
from django.core.asgi import get_asgi_application
from django.core.handlers.asgi import ASGIHandler

try:
  import uvicorn
except ImportError:
  uvicorn = None

from pistoke.kasittelija import WebsocketKasittelija


if 'django.contrib.staticfiles' in settings.INSTALLED_APPS:
  # Mikäli `staticfiles` on käytössä,
  # (1) periytetään `runserver`-komento
  #     kyseisessä sovelluksessa periytetystä `runserver`-komennosta ja
  # (2) käytetään erillistä ASGI-käsittelijää
  #     staattisten tiedostojen tarjoilemiseen.
  from django.contrib.staticfiles.management.commands.runserver import (
    Command as _Command
  )
  from django.contrib.staticfiles.handlers import StaticFilesHandlerMixin
  class static_handler(StaticFilesHandlerMixin, ASGIHandler):
    ''' Vrt. staticfiles.handlers.ASGIStaticFilesHandler '''
    def __init__(self):
      # pylint: disable=super-init-not-called
      # Ei kutsuta `ASGIHandler.__init__`-metodia,
      # joka lataa välikkeet (middleware); niitä ei tarvita.
      self.base_url = urlparse(self.get_base_url())
    # class static_handler
  # if 'django.contrib.staticfiles' in settings.INSTALLED_APPS

else:
  # Muuten periytetään suoraan django.core-versiosta.
  from django.core.management.commands.runserver import (
    Command as _Command
  )
  static_handler = None
  # else


class Kasittelija:
  ''' Käsittele tulevat ASGI-pyynnöt oikean käsittelijän läpi. '''

  def __init__(self, django, websocket, static=False):
    ''' Alusta tyyppikohtaiset käsittelijät. '''
    self.django = django
    self.websocket = websocket
    self.static = static_handler() if static_handler and static else None
    # def __init__

  async def __call__(self, scope, receive, send):
    ''' Välitä pyyntö oikealle käsittelijälle otsaketietojen mukaan. '''
    # pylint: disable=not-callable, protected-access
    if scope['type'] == 'http':
      if self.static and self.static._should_handle(scope['path']):
        return await self.static(scope, receive, send)
      else:
        return await self.django(scope, receive, send)
    elif scope['type'] == 'websocket':
      return await self.websocket(scope, receive, send)
    else:
      raise ValueError(f'tuntematon pyyntö: {scope["type"]}')
    # async def __call__

  # class Kasittelija


# Luodaan kaksi erillistä ASGI-käsittelyfunktiota:
# staticfilesillä ja ilman sitä.
uvicorn_application = Kasittelija(
  django=get_asgi_application(),
  websocket=WebsocketKasittelija(),
)
uvicorn_application_static = Kasittelija(
  django=get_asgi_application(),
  websocket=WebsocketKasittelija(),
  static=True,
)


class Command(_Command):
  def add_arguments(self, parser):
    super(Command, self).add_arguments(parser)
    if uvicorn is not None:
      # Mikäli uvicorn on asennettu, käytetään oletuksena
      # ASGI-palvelinta ja lisätään vipu WSGI:n käyttöön.
      parser.add_argument(
        '--wsgi',
        action='store_false',
        dest='asgi',
        default=True,
        help='Käynnistä WSGI-palvelin (oletus ASGI).'
      )
    # def add_arguments

  def asgi_run(self, addr, port, asgi_handler, **options):
    if uvicorn is None:
      raise RuntimeError(
        'Uvicorn-paketti puuttuu.'
        ' ASGI-testipalvelimen ajaminen edellyttää tätä.'
        'Asenna komennolla `pip install uvicorn`.'
      )
    uvicorn.run(
      asgi_handler,
      host=addr,
      port=port,
      log_level={
        0: 'critical', 1: 'error', 2: 'info', 3: 'debug'
      }.get(options['verbosity'], 1),
      reload=options['use_reloader'],
    )
    # def run

  def run(self, **options):
    if options.get('asgi', False) and options['use_reloader']:
      # Ohita mahdollinen automaattinen uudelleenlataus tässä.
      # Käytetään `uvicorn --reload`-parametriä.
      self.inner_run(None, **options)
    else:
      super().run(**options)
    # def run

  class ASGIPalvelin(Exception):
    ''' Poikkeusluokka ASGI-käsittelyn ilmaisemiseksi. '''

  def inner_run(self, *args, **options):
    '''
    Mikäli super() nostaa em. poikkeuksen,
    suoritetaan ASGI-käsittelijä (Uvicorn).
    '''
    try:
      super().inner_run(*args, **options)
    except self.ASGIPalvelin:
      # Poimitaan haluttu sovellus käsillä olevasta moduulista.
      handler = self.__class__.__module__ + (
        ':uvicorn_application_static'
        if options['use_static_handler'] and (
          settings.DEBUG or options['insecure_serving']
        )
        else ':uvicorn_application'
      )
      self.asgi_run(self.addr, int(self.port), handler, **options)
    # def inner_run

  def get_handler(self, *args, **options):
    if options.get('asgi', False):
      # Katkaistaan `super().inner_run`-metodin suoritus ennen
      # WSGI-palvelimen käynnistystä.
      raise self.ASGIPalvelin
    return super().get_handler(*args, **options)
    # def get_handler

  # class Command
